import functools as ft
from collections.abc import Callable
from typing import Any, Optional, Union

import jax
import jax._src.pretty_printer as pp
import jax._src.source_info_util as source_info_util
import jax.core
import jax.interpreters.ad as ad
import jax.interpreters.batching as batching
import jax.interpreters.mlir as mlir
import jax.numpy as jnp
import jax.tree_util as jtu
import numpy as np
from jaxtyping import Array, Int, PyTree

from .._ad import filter_jvp
from .._caches import internal_lru_caches
from .._compile_utils import hashable_combine, hashable_partition
from .._eval_shape import filter_eval_shape
from .._filters import combine, is_array, partition
from .._module import field, Module, module_update_wrapper
from .._vmap_pmap import filter_vmap
from . import _primitive
from ._primitive import (
    filter_primitive_batching,
    filter_primitive_bind,
    filter_primitive_def,
    filter_primitive_jvp,
    filter_primitive_transpose,
    materialise_zeros,
)


def _is_shapedarray(x):
    return isinstance(x, jax.core.ShapedArray)


def _to_shapedarray(x):
    if isinstance(x, jax.ShapeDtypeStruct):
        return jax.core.ShapedArray(x.shape, x.dtype)
    else:
        return x


def _only_shapedarrays(rule):
    def _rule_wrapper(*args, **params):
        for arg in args:
            if not _is_shapedarray(arg):
                raise NotImplementedError(
                    "noinline only supports ShapedArrays in abstract evaluation"
                )
        return rule(*args, **params)

    return _rule_wrapper


@ft.lru_cache(maxsize=128)
def _cache_filter_eval_shape(key):
    flat_dynamic, treedef_dynamic, static = key
    flat_dynamic = [jax.ShapeDtypeStruct(shape, dtype) for shape, dtype in flat_dynamic]
    dynamic = jtu.tree_unflatten(treedef_dynamic, flat_dynamic)
    (
        abstract_fn,
        transforms,
        args,
    ) = hashable_combine(dynamic, static)
    assert transforms[0] is _impl_transform
    abstract_fn = _abstract_transform(abstract_fn)
    for transform in transforms[1:]:
        abstract_fn = transform(abstract_fn)
    return filter_eval_shape(abstract_fn, args)


internal_lru_caches.append(_cache_filter_eval_shape)


def _is_undefined(x):
    return type(x) is ad.UndefinedPrimal


def _is_none(x):
    return x is None


def _is_not_mapped(x):
    return x is batching.not_mapped


def _move_to_front(input, batch_axis):
    if batch_axis is batching.not_mapped:
        return input
    else:
        return jnp.swapaxes(input, 0, batch_axis)


def _int_to_zero(batch_axis):
    if batch_axis is batching.not_mapped:
        return batch_axis
    else:
        return 0


@ft.lru_cache(maxsize=None)
def _get_callback(treedef, static, is_float0):
    @ft.partial(jax.jit, static_argnums=0)
    def callback(static_fn, dynamic):
        # pure_callback casts float0 to bool, so need to cast back
        assert len(dynamic) == len(is_float0)
        dynamic = [
            np.broadcast_to(np.zeros((), dtype=jax.dtypes.float0), x.shape) if y else x
            for x, y in zip(dynamic, is_float0)
        ]
        iter_dynamic = iter(dynamic)
        flat = [next(iter_dynamic) if x is None else x for x in static]
        assert next(iter_dynamic, None) is None
        transforms, args = jtu.tree_unflatten(treedef, flat)
        for transform in transforms:
            static_fn = transform(static_fn)
        out = static_fn(args)
        return tuple(jtu.tree_leaves(out))

    def callback_lookup(dynamic_index, *dynamic):
        static_fn = _index_to_fn[dynamic_index.item()]
        return callback(static_fn, dynamic)

    return callback_lookup


internal_lru_caches.append(_get_callback)


def _impl_transform(static_fn):
    def _impl_transform_impl(args):
        dynamic_fn, args, kwargs = args
        fn = hashable_combine(dynamic_fn, static_fn)
        return fn(*args, **kwargs)

    return _impl_transform_impl


def _abstract_transform(abstract_fn):
    def _abstract_transform_impl(args):
        dynamic_fn, args, kwargs = args
        return abstract_fn(dynamic_fn, *args, **kwargs)

    return _abstract_transform_impl


def _jvp_transform(static_fn):
    def _jvp_transform_impl(args):
        args, t_args = args
        tang_is_none = jtu.tree_map(_is_none, t_args, is_leaf=_is_none)
        no_tang, has_tang = partition(args, tang_is_none, is_leaf=_is_none)

        def _to_jvp(_has_tang):
            _args = combine(_has_tang, no_tang)
            return static_fn(_args)

        _, t_out = filter_jvp(_to_jvp, (has_tang,), (t_args,))
        return t_out

    return _jvp_transform_impl


class _MetaTransposeTransform(Module):
    undefined: PyTree[jax.core.ShapedArray]

    def __call__(self, static_fn):
        def _transpose_transform_impl(args):
            defined, cts_out = args

            def _to_transpose(_undefined):
                _args = combine(defined, _undefined)
                return static_fn(_args)

            (cts_undefined,) = jax.linear_transpose(_to_transpose, self.undefined)(
                cts_out
            )
            return cts_undefined

        return _transpose_transform_impl


class _MetaBatchTransform(Module):
    batch_axes: PyTree[Union[batching.NotMapped, int]]

    def __call__(self, static_fn):
        return filter_vmap(static_fn, in_axes=(self.batch_axes,))


@filter_primitive_def
def _noinline_impl(dynamic_index, abstract_fn, transforms, args):
    del abstract_fn
    static_fn = _index_to_fn[dynamic_index.item()]
    for transform in transforms:
        static_fn = transform(static_fn)
    return static_fn(args)


@_only_shapedarrays
@filter_primitive_def
def _noinline_abstract(dynamic_index, abstract_fn, transforms, args):
    del dynamic_index
    dynamic, static = hashable_partition(
        (abstract_fn, transforms, args), _is_shapedarray
    )
    flat_dynamic, treedef_dynamic = jtu.tree_flatten(dynamic)
    key = tuple((x.shape, x.dtype) for x in flat_dynamic), treedef_dynamic, static
    out_struct = _cache_filter_eval_shape(key)
    return jtu.tree_map(_to_shapedarray, out_struct)


@filter_primitive_jvp
def _noinline_jvp(primals, tangents):
    # TODO: add custom partial-eval rule to avoid the double-noinline?
    dynamic_index, abstract_fn, transforms, args = primals
    t_dynamic_index, t_abstract_fn, t_transforms, t_args = tangents
    assert (
        len(jtu.tree_leaves((t_dynamic_index, t_abstract_fn, t_transforms))) == 0
    )  # all none
    del t_dynamic_index, t_abstract_fn, t_transforms
    tangents = jtu.tree_map(materialise_zeros, args, t_args, is_leaf=_is_none)
    primal_outs = filter_primitive_bind(noinline_p, *primals)
    tangent_outs = filter_primitive_bind(
        noinline_p,
        dynamic_index,
        abstract_fn,
        transforms + [_jvp_transform],
        (args, t_args),
    )
    return primal_outs, tangent_outs


@filter_primitive_transpose(materialise_zeros=True)  # pyright: ignore
def _noinline_transpose(inputs, cts_out):
    dynamic_index, abstract_fn, transforms, args = inputs
    assert all(
        not _is_undefined(x)
        for x in jtu.tree_leaves(
            (dynamic_index, abstract_fn, transforms), is_leaf=_is_undefined
        )
    )
    # Note that `defined` may also include non-JAX-arrays
    undefined, defined = partition(args, _is_undefined, is_leaf=_is_undefined)
    undefined = jtu.tree_map(lambda x: x.aval, undefined, is_leaf=_is_undefined)
    cts_args = filter_primitive_bind(
        noinline_p,
        dynamic_index,
        abstract_fn,
        transforms + [_MetaTransposeTransform(undefined)],
        (defined, cts_out),
    )
    cts_rest = jtu.tree_map(lambda _: None, (dynamic_index, abstract_fn, transforms))
    return cts_rest + (cts_args,)


@filter_primitive_batching
def _noinline_batch(inputs, batch_axes):
    dynamic_index, abstract_fn, transforms, args = inputs
    dynamic_index_bdim, abstract_fn_bdim, transforms_bdim, args_bdim = batch_axes
    assert len(jtu.tree_leaves((abstract_fn_bdim, transforms_bdim))) == 0  # all none
    if dynamic_index_bdim is not batching.not_mapped:
        # The batch rule for `lax.cond` with vmap'd predicate simply
        # broadcasts all constants in the branches. In particular it may broadcast
        # this. We simply need to ignore this and return to having a single dynamic
        # index.
        # This is actually a silent error if you do something exceptionally silly, and
        # try to manually combine two different `noinline`d functions. There's no real
        # way to catch this that wouldn't slow things down at runtime though, I don't
        # think.
        assert jnp.ndim(dynamic_index) == 1
        dynamic_index = dynamic_index[0]
    del dynamic_index_bdim, abstract_fn_bdim, transforms_bdim
    args = jtu.tree_map(_move_to_front, args, args_bdim, is_leaf=_is_not_mapped)
    args_bdim = jtu.tree_map(_int_to_zero, args_bdim, is_leaf=_is_not_mapped)
    out = filter_primitive_bind(
        noinline_p,
        dynamic_index,
        abstract_fn,
        transforms + [_MetaBatchTransform(args_bdim)],
        args,
    )
    return out, jtu.tree_map(lambda _: 0, out)


def _pp_transform(x):
    if x is _jvp_transform:
        return "jvp"
    elif type(x) is _MetaTransposeTransform:
        return "transpose"
    elif type(x) is _MetaBatchTransform:
        return "vmap"
    else:
        assert False


def _noinline_pretty_print(eqn, context, settings):
    _, abstract_fn, transforms, _ = jtu.tree_unflatten(
        eqn.params["treedef"], eqn.params["static"]
    )
    pretty_params = dict(abstract_fn=abstract_fn)
    if type(eqn.invars[0]) is jax.core.Literal:
        static_fn_leaves, static_fn_treedef = _index_to_fn[eqn.invars[0].val]
        pretty_params["static_fn"] = jtu.tree_unflatten(
            static_fn_treedef, static_fn_leaves
        )
    if len(transforms) > 1:
        # skip impl
        transforms = [_pp_transform(x) for x in transforms[1:]]
        pretty_params["transforms"] = transforms
    lhs = jax.core.pp_vars(eqn.outvars, context, print_shapes=settings.print_shapes)
    rhs = [
        pp.text(eqn.primitive.name),
        jax.core.pp_kv_pairs(sorted(pretty_params.items()), context, settings),
        pp.text(" ") + jax.core.pp_vars(eqn.invars, context),
    ]
    annotation = (
        source_info_util.summarize(eqn.source_info) if settings.source_info else None
    )
    return pp.concat([lhs, pp.text(" = ", annotation=annotation), *rhs])


# Not a PyTree
class _MlirWrapper:
    def __init__(self, val):
        self.val = val


def _noinline_mlir(ctx, *dynamic, treedef, static, flatten, **kwargs):
    assert len(kwargs) == 0
    assert flatten.called()
    dynamic = [_MlirWrapper(x) for x in dynamic]
    abstract_dynamic = [_MlirWrapper(x) for x in ctx.avals_in]
    # This is really an internal implementation detail of another component that we're
    # messing with here.
    # Fortunately `noinline` and `primitive` are both inside Equinox, so this isn't too
    # bad.
    flat = _primitive._combine(dynamic, static)
    abstract_flat = _primitive._combine(abstract_dynamic, static)
    index, _, transforms, args = jtu.tree_unflatten(treedef, flat)
    abstract_index, _, abstract_transforms, abstract_args = jtu.tree_unflatten(
        treedef, abstract_flat
    )
    flat, treedef = jtu.tree_flatten((transforms, args))
    abstract_flat, abstract_treedef = jtu.tree_flatten(
        (abstract_transforms, abstract_args)
    )
    assert treedef == abstract_treedef
    dynamic = [x.val for x in flat if type(x) is _MlirWrapper]
    abstract_dynamic = [x.val for x in abstract_flat if type(x) is _MlirWrapper]
    static = tuple(None if type(x) is _MlirWrapper else x for x in flat)
    is_float0 = tuple(x.dtype == jax.dtypes.float0 for x in abstract_dynamic)
    callback = _get_callback(treedef, static, is_float0)

    vals_in = [index.val] + dynamic
    avals_in = [abstract_index.val] + abstract_dynamic
    result, _, keepalive = mlir.emit_python_callback(
        ctx,
        callback,
        None,
        vals_in,
        avals_in,
        ctx.avals_out,
        False,
        sharding=None,
    )
    ctx.module_context.add_keepalive(keepalive)
    return result


noinline_p = jax.core.Primitive("noinline")
noinline_p.multiple_results = True
noinline_p.def_impl(_noinline_impl)
noinline_p.def_abstract_eval(_noinline_abstract)
ad.primitive_jvps[noinline_p] = _noinline_jvp
ad.primitive_transposes[noinline_p] = _noinline_transpose
batching.primitive_batchers[noinline_p] = _noinline_batch
jax.core.pp_eqn_rules[noinline_p] = _noinline_pretty_print
mlir.register_lowering(noinline_p, _noinline_mlir)


_fn_to_index = {}
_index_to_fn = []


class _NoInlineWrapper(Module):
    dynamic_index: Int[Array, ""]
    abstract_fn: Callable = field(static=True)
    dynamic_fn: Any

    @property
    def __wrapped__(self):
        return self.abstract_fn

    def __call__(self, *args, **kwargs):
        return filter_primitive_bind(
            noinline_p,
            self.dynamic_index,
            self.abstract_fn,
            [_impl_transform],
            (self.dynamic_fn, args, kwargs),
        )


def noinline(
    fn: Callable, abstract_fn: Optional[Callable] = None  # pyright: ignore
) -> Callable:
    """Marks a function as not being inlined into a larger computation.
    This can help to reduce compile time at the expense of increased runtime.

    `fn` can be any callable `PyTree[Any] -> PyTree[Any]`. In addition `fn`
    itself may be any (callable) PyTree; any JAX arrays in `fn`'s PyTree are
    also treated as inputs to the noinline'd computation.

    `abstract_fn` determines the shapes/dtypes/pytrees of the output. (And must
     return results consistent with `fn`.) If `fn` is called as
    `fn(*args, **kwargs)`, then `abstract_fn` is called as
    ```python
    eqx.filter_eval_shape(
        abstract_fn, eqx.filter(fn, eqx.is_array), *args, **kwargs)
    )
    ```
    If not passed then `abstract_fn` is automatically constructed from `fn`.
    Specifying it is useful as noinline'd functions sharing the same
    `abstract_fn` may be substituted for each other without recompiling the
    enclosing computation graph; see the second example below.

    !!! Example

        ```python
        @noinline
        def f(x, y):
            print("Tracing!")
            return x + y

        @jax.jit
        def g(x, y):
            a = f(x, jnp.array(1))
            b = f(x, jnp.array(2))
            c = f(y, jnp.array(1))
            d = f(y, jnp.array(2))
            return a + b + c + d

        g(1, 1)
        ```

        In this example, `Tracing!` is printed twice. Once just for abstract
        evaluation (only) to figure out out its return shape in `f`'s
        computation graph. (Which is computationally cheap.) And then once
        when it is actually compiled as part of its own computation graph.
        (Which is computationally expensive.)

        Without the `noinline`, we would see `Tracing!` printed four times:
        once for each call, and with it being compiled each time. (Every time
        being computationally expensive.)

        Note how the `1` and the `2` are wrapped in arrays. If they were still
        Python scalars then they would be treated as static arguments, which
        will lead to recompilations of `f`. (E.g. what if there was an
        `if y == 1` command in there?)

    !!! Example

        ```python
        def abstract(_, x, y):
            return jnp.broadcast_arrays(x, y)[0]

        def f(x, y):
            print("Compiling f!")
            return x + y

        def g(x, y):
            print("Compiling g!")
            return x * y

        f = noinline(f, abstract)
        g = noinline(g, abstract)

        @jax.jit
        def call(fn, x, y):
            print("Compiling call!")
            return fn(x, y)

        call(f, 1, 1)  # Compiling call! Compiling f!
        call(g, 1, 1)  # Compiling g!
        ```

        In this example, we see how noinline'd functions with the same
        `abstract_fn` may be passed as inputs to the main call graph,
        and swapped without needing to recompile the main call graph.

    """
    dynamic_fn, static_fn = hashable_partition(fn, is_array)
    if abstract_fn is None:

        def abstract_fn(__dynamic_fn, *args, **kwargs):
            _fn = hashable_combine(__dynamic_fn, static_fn)
            return _fn(*args, **kwargs)

    try:
        dynamic_index = _fn_to_index[static_fn]
    except KeyError:
        dynamic_index = len(_index_to_fn)
        _fn_to_index[static_fn] = dynamic_index
        _index_to_fn.append(static_fn)
    dynamic_index = jnp.array(dynamic_index)
    noinline_fn = _NoInlineWrapper(dynamic_index, abstract_fn, dynamic_fn)
    return module_update_wrapper(noinline_fn, abstract_fn)



from collections.abc import Callable, Sequence
from typing import Any, Literal, Optional, TypeVar, Union

import jax
import jax.lax as lax
import jax.numpy as jnp
import jax.tree_util as jtu
from jaxtyping import Array, Bool

from .bounded import bounded_while_loop
from .checkpointed import checkpointed_while_loop
from .common import common_rewrite


_Carry = TypeVar("_Carry")
_X = TypeVar("_X")
_Y = TypeVar("_Y")
_Bool = Union[bool, Bool[Array, ""]]
_Node = Any


def while_loop(
    cond_fun: Callable[[_Carry], _Bool],
    body_fun: Callable[[_Carry], _Carry],
    init_val: _Carry,
    *,
    max_steps: Optional[int] = None,
    buffers: Optional[Callable[[_Carry], Union[_Node, Sequence[_Node]]]] = None,
    kind: Literal["lax", "checkpointed", "bounded"],
    checkpoints: Optional[int] = None,
    base: int = 16,
) -> _Carry:
    """A better while loop, supporting (1) reverse-mode autodifferentiation; (2) online
    checkpointing schemes; (3) efficient in-place scatters (normally XLA tends to make
    unnecessary copies).

    **Arguments:**

    - `cond_fun`: As `lax.while_loop`.
    - `body_fun`: As `lax.while_loop`.
    - `init_val`: As `lax.while_loop`.

    - `max_steps`: A bound on the maximum number of steps, after which the loop
        terminates unconditionally. Can be set to `None` for arbitrarily many steps.

    - `buffers`: If passed, then every leaf of `tree_leaves(buffers(init_val))` must
        be an array; all such arrays become buffers supporting only `[]` and
        `.at[].set()`. However they will act efficiently, without spurious copies.
        You should avoid performing in-place updates to any quantity that is not a
        buffer.

    - `kind`: The type of while loop that is lowered to underneath. This may either be
        `"lax"`, `"checkpointed"`, or `"bounded"`.

        If `kind` is `"lax"` then the loop is efficiently forward-mode
        autodifferentiable, but does not support reverse-mode autodifferentiation.

        If `kind` is `"checkpointed"` then the loop is efficiently reverse-mode
        autodifferentiable, but does not support forward-mode autodifferentiation, and
        all `buffers` will be write-only whilst inside the loop.

        If `kind` is `"bounded"` then the loop may be both forward- and reverse-mode
        autodifferentiated, but it requires an `int` value for `max_steps`, and as
        `max_steps` grows then time and memory usage will increase.

    - `checkpoints`: Only used if `kind="checkpointed"`. Specifies the number of
        checkpoints to use; if `None` then this is automatically derived from
        `max_steps`.

    - `base`: Only used if `kind="bounded"`. Run time will increase slightly as `base`
        increases. Compilation time will decrease substantially as
        `math.ceil(math.log(max_steps, base))` decreases. (Which happens as `base`
        increases.)

    !!! Danger

        Note that `buffers` is subject to the following restrictions:

        - You should never write to the same location twice.
        - You should only read from it (`buf[i]`) at locations (`i`) that you have
          written to previously.

        These assumptions are *completely unchecked* and you will get incorrect
        gradients if you violate these assumptions.

    **Returns:**

    The final value; as `lax.while_loop`.
    """

    if kind == "lax":
        del kind, checkpoints, base
        cond_fun_, body_fun_, init_val_, _ = common_rewrite(
            cond_fun, body_fun, init_val, max_steps, buffers
        )
        del cond_fun, body_fun, init_val
        _, _, final_val = lax.while_loop(cond_fun_, body_fun_, init_val_)
        return final_val
    elif kind == "checkpointed":
        del kind, base
        return checkpointed_while_loop(
            cond_fun,
            body_fun,
            init_val,
            max_steps=max_steps,
            buffers=buffers,
            checkpoints=checkpoints,
        )
    elif kind == "bounded":
        del kind, checkpoints
        if max_steps is None:
            raise ValueError("kind='bounded' requires `max_steps` to be specified")
        return bounded_while_loop(
            cond_fun,
            body_fun,
            init_val,
            max_steps=max_steps,
            buffers=buffers,
            base=base,
        )
    else:
        raise ValueError(f"Unrecognised kind of while loop '{kind}'")


def scan(
    f: Callable[[_Carry, _X], tuple[_Carry, _Y]],
    init: _Carry,
    xs: _X,
    length: Optional[int] = None,
    *,
    buffers: Optional[Callable[[_Carry], Union[_Node, Sequence[_Node]]]] = None,
    kind: Literal["lax", "checkpointed"],
    checkpoints: Union[None, int, Literal["all"]] = None,
) -> tuple[_Carry, _Y]:
    """As `jax.lax.scan`, but with optional checkpointing to reduce memory usage.

    **Arguments:**

    - `f`: As `jax.lax.scan`.
    - `init`: As `jax.lax.scan`.
    - `xs`: As `jax.lax.scan`.
    - `length`: As `jax.lax.scan`.

    - `buffers`: If passed, then every leaf of `tree_leaves(buffers(init))` must
        be an array; all such arrays become buffers supporting only `[]` and
        `.at[].set()`. However they will act efficiently, without spurious copies.
        You should avoid performing in-place updates to any quantity that is not a
        buffer.

    - `kind`: The type of scan that is lowered to underneath. This may either be
        `"lax"` or `"checkpointed"`.

        If `kind` is `"lax"` then the usual `lax.scan` is used.

        If `kind` is `"checkpointed"` then the scan uses checkpointing to reduce memory
        usage. It will not be forward-mode autodifferentiable.

    - `checkpoints`: Only used if `kind="checkpointed"`. Specifies the number of
        checkpoints to use; if `None` then this is set proportional to `sqrt(length)`.
        Can also be a string `"all"`, representing checkpointing every step.

    Returns:

    As `jax.lax.scan`.
    """

    init, xs = jtu.tree_map(jnp.asarray, (init, xs))

    if kind == "lax":
        return lax.scan(f, init, xs, length)

    lengths = {jnp.shape(x)[0] for x in jtu.tree_leaves(xs)}
    if length is not None:
        lengths.add(length)
    if len(lengths) == 1:
        length = lengths.pop()
    else:
        raise ValueError(f"Got inconsistent lengths in scan: {lengths}.")
    del lengths

    if checkpoints == "all":
        checkpoints = length

    def cond_fun(val):
        return True

    def body_fun(val):
        i, carry, ys = val
        x = jtu.tree_map(lambda z: z[i], xs)
        carry, y = f(carry, x)
        ys = jtu.tree_map(lambda z, zs: zs.at[i].set(z), y, ys)
        return i + 1, carry, ys

    x0 = jtu.tree_map(lambda z: z[0], xs)
    _, y0_shape = jax.eval_shape(f, init, x0)
    ys = jtu.tree_map(lambda z: jnp.empty((length,) + z.shape, z.dtype), y0_shape)
    init_val = (0, init, ys)

    if buffers is None:
        _buffers = lambda val: val[2]
    else:
        node_or_nodes = buffers(init)
        is_tuple = True
        is_leaf = lambda node: node is node_or_nodes

        def _find(node):
            nonlocal is_tuple
            if is_leaf(node):
                is_tuple = False

        jtu.tree_map(_find, init, is_leaf=is_leaf)
        if is_tuple:
            _buffers = lambda val: tuple(buffers(val[1])) + (val[2],)
        else:
            _buffers = lambda val: (buffers(val[1]), val[2])

    _, carry, ys = checkpointed_while_loop(
        cond_fun,
        body_fun,
        init_val,
        buffers=_buffers,
        max_steps=length,
        checkpoints=checkpoints,
    )
    return carry, ys



internal_caches = []
internal_lru_caches = []


def clear_caches():
    """Clears internal Equinox caches.

    Best used before calling `jax.clear_caches()` and `jax.clear_backends()`.

    **Arguments:**

    None.

    **Returns:**

    None.
    """
    for cache in internal_caches:
        cache.clear()
    for cache in internal_lru_caches:
        cache.cache_clear()



from collections.abc import Callable
from typing import Any

import jax
import jax.core
import jax.interpreters.ad as ad
import jax.interpreters.batching as batching
import jax.interpreters.mlir as mlir
import jax.numpy as jnp
import jax.tree_util as jtu

from .._ad import filter_custom_vjp
from .._errors import error_if
from .._filters import combine, filter, is_array, is_array_like, partition
from .._module import Module
from .._pretty_print import tree_pformat


def announce_transform(
    x, name=None, intermediates=False, announce: Callable[[str], Any] = print
):
    """Identity function on an arbitrary PyTree. Announces each time it is parsed as
    part of a jaxpr.
    """
    if name is None:
        name = ""
    else:
        name = f"({name}) "
    array, nonarray = partition(x, is_array)
    flat, treedef = jtu.tree_flatten(array)
    flat = announce_jaxpr_p.bind(
        *flat, stack=(), name=name, intermediates=intermediates, announce=announce
    )
    array = jtu.tree_unflatten(treedef, flat)
    return combine(array, nonarray)


def _impl(*x, stack, name, intermediates, announce):
    del intermediates
    stack = stack + ("impl,")
    announce(name + ":".join(stack))
    return x


def _abstract(*x, stack, name, intermediates, announce):
    del intermediates
    stack = stack + ("abstract",)
    announce(name + ":".join(stack))
    return x


def _jvp(p, t, *, stack, name, intermediates, announce):
    if intermediates:
        _stack = stack + ("jvp",)
        announce(name + ":".join(_stack))
    p_stack = stack + ("jvp_p",)
    t_stack = stack + ("jvp_t",)
    p_out = announce_jaxpr_p.bind(
        *p, stack=p_stack, name=name, intermediates=intermediates, announce=announce
    )
    t_out = announce_jaxpr_p.bind(
        *t, stack=t_stack, name=name, intermediates=intermediates, announce=announce
    )
    return p_out, t_out


def _transpose(ct, p, *, stack, name, intermediates, announce):
    stack = stack + ("transpose",)
    if intermediates:
        announce(name + ":".join(stack))
    return announce_jaxpr_p.bind(
        *ct, stack=stack, name=name, intermediates=intermediates, announce=announce
    )


def _batching(p, b, *, stack, name, intermediates, announce):
    stack = stack + ("vmap",)
    if intermediates:
        announce(name + ":".join(stack))
    out = announce_jaxpr_p.bind(
        *p, stack=stack, name=name, intermediates=intermediates, announce=announce
    )
    return out, b


def _mlir(*x, stack, name, intermediates, announce):
    del intermediates
    stack = stack + ("mlir",)
    announce(name + ":".join(stack))
    return x


announce_jaxpr_p = jax.core.Primitive("announce_jaxpr")
announce_jaxpr_p.multiple_results = True
announce_jaxpr_p.def_impl(_impl)
announce_jaxpr_p.def_abstract_eval(_abstract)
ad.primitive_jvps[announce_jaxpr_p] = _jvp
ad.primitive_transposes[announce_jaxpr_p] = _transpose
batching.primitive_batchers[announce_jaxpr_p] = _batching
mlir.register_lowering(announce_jaxpr_p, mlir.lower_fun(_mlir, multiple_results=True))


def debug_backward_nan(x, name=None, terminate=True):
    """Debug NaNs that only occur on the backward pass.

    **Arguments:**

    - `x`: a variable to intercept.
    - `name`: an optional name to appear in printed debug statements.
    - `terminate`: whether to halt the computation if a NaN cotangent is found.

    **Returns:**

    The `x` argument is returned unchanged.

    As a side-effect, both the primal and the cotangent for `x` will be printed out
    during the backward pass.
    """
    return _debug_backward_nan(x, name, terminate)


@filter_custom_vjp
def _debug_backward_nan(x, name, terminate):
    return x


def _debug_backward_nan_fwd(x, name, terminate):
    return debug_backward_nan(x, name, terminate), None


class _LongRepr(Module):
    obj: Any

    def __repr__(self):
        return tree_pformat(self.obj, short_arrays=False)


def _debug_backward_nan_bwd(_, grad_x, x, name, terminate):
    msg = "   primals={x}\ncotangents={grad_x}"
    if name is not None:
        msg = f"{name}:\n" + msg
    jax.debug.print(  # pyright: ignore
        msg, x=_LongRepr(x), grad_x=_LongRepr(grad_x), ordered=True
    )
    if terminate:
        nans = [
            jnp.isnan(a).any() for a in jtu.tree_leaves(filter(grad_x, is_array_like))
        ]
        grad_x = error_if(grad_x, jnp.any(jnp.stack(nans)), "Encountered NaN")
    return grad_x


_debug_backward_nan.defvjp(_debug_backward_nan_fwd, _debug_backward_nan_bwd)



from collections.abc import Callable, Sequence
from typing import Any, TypeVar

import jax.lax as lax
import jax.numpy as jnp
import jax.tree_util as jtu
from jaxtyping import Array, PyTree, Scalar

from .._eval_shape import filter_eval_shape
from .._filters import is_array
from ._nontraceable import nonbatchable


# Old; deprecated
class ContainerMeta(type):
    reverse_lookup: dict

    def __new__(cls, name, bases, dict):
        assert "reverse_lookup" not in dict
        _dict = {}
        reverse_lookup = []
        i = 0
        for key, value in dict.items():
            if key.startswith("__") and key.endswith("__"):
                _dict[key] = value
            else:
                _dict[key] = i
                reverse_lookup.append(value)
                i += 1
        _dict["reverse_lookup"] = reverse_lookup
        return super().__new__(cls, name, bases, _dict)

    def __instancecheck__(cls, instance):
        if is_array(instance):
            return instance.shape == () and jnp.issubdtype(instance.dtype, jnp.integer)
        else:
            return isinstance(instance, int) or super().__instancecheck__(instance)

    def __getitem__(cls, item):
        return cls.reverse_lookup[item]

    def __len__(cls):
        return len(cls.reverse_lookup)


_X = TypeVar("_X")


def scan_trick(fn: Callable, intermediates: Sequence[Callable], init: _X) -> _X:
    def body(carry, step):
        out = fn(carry)
        step = nonbatchable(step)
        out = lax.switch(step, intermediates, out)
        return out, None

    intermediates = list(intermediates) + [lambda x: x]
    out, _ = lax.scan(body, init, xs=jnp.arange(len(intermediates)))
    return out


def eval_empty(fn: Callable, *inputs: PyTree[Any]) -> PyTree[Array]:
    out = filter_eval_shape(fn, *inputs)
    return jtu.tree_map(lambda x: jnp.empty(x.shape, x.dtype), out)


def eval_zero(fn: Callable, *inputs: PyTree[Any]) -> PyTree[Array]:
    out = filter_eval_shape(fn, *inputs)
    return jtu.tree_map(lambda x: jnp.zeros(x.shape, x.dtype), out)


def eval_full(fn: Callable, *inputs: PyTree[Any], fill_value: Scalar) -> PyTree[Any]:
    out = filter_eval_shape(fn, *inputs)
    return jtu.tree_map(lambda x: jnp.full(x.shape, fill_value, x.dtype), out)



from typing import Optional

import jax.random as jrandom
from jaxtyping import Array, Float, PRNGKeyArray

from .._module import field, Module


class Embedding(Module):
    """A simple lookup table that stores embeddings of a fixed size."""

    num_embeddings: int = field(static=True)
    embedding_size: int = field(static=True)
    weight: Array

    def __init__(
        self,
        num_embeddings: int,
        embedding_size: int,
        weight: Optional[Float[Array, "num_embeddings embedding_size"]] = None,
        *,
        key: PRNGKeyArray,
        **kwargs,
    ):
        """**Arguments:**

        - `num_embeddings`: Size of embedding dictionary.
        - `embedding_size`: Size of each embedding vector.
        - `weight`: If given, the embedding lookup table. Will be generated randomly
            if not provided.
        - `key`: A `jax.random.PRNGKey` used to provide randomness for parameter
            initialisation. (Keyword only argument.)
        """
        super().__init__(**kwargs)
        if weight is None:
            self.weight = jrandom.normal(key, (num_embeddings, embedding_size))
        else:
            if weight.shape != (num_embeddings, embedding_size):
                raise ValueError(
                    "weight must have shape (num_embeddings, embedding_size)."
                )
            self.weight = weight
        self.num_embeddings = num_embeddings
        self.embedding_size = embedding_size

    def __call__(self, x: Array, *, key: Optional[PRNGKeyArray] = None) -> Array:
        """**Arguments:**

        - `x`: The table index.
        - `key`: Ignored; provided for compatibility with the rest of the Equinox API.
            (Keyword only argument.)

        **Returns:**

        A JAX array of shape `(embedding_size,)`, from the x-th index of the embedding
        table.
        """
        return self.weight[x]



"""Introduces `AbstractVar` and `AbstractClassVar` type annotations, which can be used
to mark abstract instance attributes and abstract class attributes.

Provides enhanced versions of `abc.ABCMeta` and `dataclasses.dataclass` that respect
this type annotation. (In practice the sole consumer of these is `equinox.Module`, which
is the public API. But the two pieces aren't directly related under-the-hood.)
"""
import abc
import dataclasses
from typing import ClassVar, Generic, get_args, get_origin, TYPE_CHECKING, TypeVar
from typing_extensions import dataclass_transform


_T = TypeVar("_T")


if TYPE_CHECKING:
    from typing import ClassVar as AbstractClassVar, ClassVar as AbstractVar
else:

    class AbstractVar(Generic[_T]):
        """Used to mark an abstract instance attribute, along with its type. Used as:
        ```python
        class Foo(eqx.Module):
            attr: AbstractVar[bool]
        ```

        An `AbstractVar[T]` must be overridden by an attribute annotated with
        `AbstractVar[T]`, `AbstractClassVar[T]`, `ClassVar[T]`, `T`, or a property
        returning `T`.

        This makes `AbstractVar` useful when you just want to assert that you can access
        `self.attr` on a subclass, regardless of whether it's an instance attribute,
        class attribute, property, etc.

        Attempting to instantiate a module with an unoveridden `AbstractVar` will raise
        an error.

        !!! Example

            ```python
            import equinox as eqx

            class AbstractX(eqx.Module):
                attr1: int
                attr2: AbstractVar[bool]

            class ConcreteX(AbstractX):
                attr2: bool

            ConcreteX(attr1=1, attr2=True)
            ```

        !!! Info

            `AbstractVar` does not create a dataclass field. This affects the order of
            `__init__` argments. E.g.
            ```python
            class AbstractX(Module):
                attr1: AbstractVar[bool]

            class ConcreteX(AbstractX):
                attr2: str
                attr1: bool
            ```
            should be called as `ConcreteX(attr2, attr1)`.

        ## Known issues

        Due to a Pyright bug
        ([#4965](https://github.com/microsoft/pyright/issues/4965)), this must be
        imported as:
        ```python
        if TYPE_CHECKING:
            from typing import ClassVar as AbstractVar
        else:
            from equinox.internal import AbstractVar
        ```
        """

    # We can't just combine `ClassVar[AbstractVar[...]]`. At static checking time we
    # fake `AbstractVar` as `ClassVar` to prevent it from appearing in __init__
    # signatures. This means that static type checkers think they see
    # `ClassVar[ClassVar[...]]` which is not allowed.
    class AbstractClassVar(Generic[_T]):
        """Used to mark an abstract class attribute, along with its type. Used as:
        ```python
        class Foo(eqx.Module):
            attr: AbstractClassVar[bool]
        ```

        An `AbstractClassVar[T]` can be overridden by an attribute annotated with
        `AbstractClassVar[T]`, or `ClassVar[T]`. This makes `AbstractClassVar` useful
        when you want to assert that you can access `cls.attr` on a subclass.

        Attempting to instantiate a module with an unoveridden `AbstractClassVar` will
        raise an error.

        !!! Example

            ```python
            import equinox as eqx
            from typing import ClassVar

            class AbstractX(eqx.Module):
                attr1: int
                attr2: AbstractClassVar[bool]

            class ConcreteX(AbstractX):
                attr2: ClassVar[bool] = True

            ConcreteX(attr1=1)
            ```

        !!! Info

            `AbstractClassVar` does not create a dataclass field. This affects the order
            of `__init__` argments. E.g.
            ```python
            class AbstractX(Module):
                attr1: AbstractClassVar[bool]

            class ConcreteX(AbstractX):
                attr2: str
                attr1: ClassVar[bool] = True
            ```
            should be called as `ConcreteX(attr2)`.

        ## Known issues


        Due to a Pyright bug
        ([#4965](https://github.com/microsoft/pyright/issues/4965)), this must be
        imported as:
        ```python
        if TYPE_CHECKING:
            from typing import ClassVar as AbstractClassVar
        else:
            from equinox.internal import AbstractClassVar
        ```
        """


def _process_annotation(annotation):
    is_abstract = False
    is_class = False
    bad_class_var_annotation = False
    if isinstance(annotation, str):
        if annotation.startswith("AbstractVar["):
            is_abstract = True
            is_class = False
        elif annotation.startswith("AbstractClassVar["):
            is_abstract = True
            is_class = True
        elif annotation.startswith("AbstractVar") or annotation.startswith(
            "AbstractClassVar"
        ):
            bad_class_var_annotation = True
    else:
        if get_origin(annotation) is AbstractVar:
            is_abstract = True
            is_class = False
        elif get_origin(annotation) is AbstractClassVar:
            is_abstract = True
            is_class = True
        elif annotation in (AbstractVar, AbstractClassVar):
            bad_class_var_annotation = True
    if bad_class_var_annotation:
        raise TypeError("Cannot use unsubscripted `AbstractVar` or `AbstractClassVar`.")
    return is_abstract, is_class


_sentinel = object()


def _is_concretisation(sub, super):
    if isinstance(sub, str) != isinstance(super, str):
        raise NotImplementedError(
            "Cannot mix stringified and non-stringified abstract " "type annotations"
        )
    if isinstance(sub, str):
        if super.startswith("AbstractVar["):
            return sub == super or sub == super[len("AbstractVar[") : -1]
        elif super.startswith("AbstractClassVar["):
            return sub == super or sub == super[len("AbstractClassVar[") : -1]
        else:
            assert False
    else:
        if get_origin(super) is AbstractVar:
            if get_origin(sub) in (AbstractVar, AbstractClassVar, ClassVar):
                return get_args(sub) == get_args(super)
            else:
                return (sub,) == get_args(super)
        elif get_origin(super) is AbstractClassVar:
            if get_origin(sub) in (AbstractClassVar, ClassVar):
                return get_args(sub) == get_args(super)
            else:
                return False
        else:
            assert False


class ABCMeta(abc.ABCMeta):
    def register(cls, subclass):
        raise ValueError

    def __new__(mcs, name, bases, namespace, /, **kwargs):
        cls = super().__new__(mcs, name, bases, namespace, **kwargs)
        abstract_vars = dict()
        abstract_class_vars = dict()
        for attr, group in [
            ("__abstractvars__", abstract_vars),
            ("__abstractclassvars__", abstract_class_vars),
        ]:
            for base in bases:
                for name, annotation in base.__dict__.get(attr, dict()).items():
                    try:
                        existing_annotation = group[name]
                    except KeyError:
                        pass
                    else:
                        if isinstance(annotation, str) != isinstance(
                            existing_annotation, str
                        ):
                            raise NotImplementedError(
                                "Cannot mix stringified and non-stringified abstract "
                                "type annotations"
                            )
                        if not (
                            _is_concretisation(annotation, existing_annotation)
                            or _is_concretisation(existing_annotation, annotation)
                        ):
                            raise TypeError(
                                "Base classes have mismatched type annotations for "
                                f"{name}"
                            )
                    if "__annotations__" in cls.__dict__:
                        try:
                            new_annotation = cls.__annotations__[name]
                        except KeyError:
                            pass
                        else:
                            if not _is_concretisation(new_annotation, annotation):
                                raise TypeError(
                                    "Base class and derived class have mismatched type "
                                    f"annotations for {name}"
                                )
                    # Not just `if name not in namespace`, as `cls.__dict__` may be
                    # slightly bigger from `__init_subclass__`.
                    if name not in cls.__dict__:
                        group[name] = annotation
        if "__annotations__" in cls.__dict__:
            for name, annotation in cls.__annotations__.items():
                is_abstract, is_class = _process_annotation(annotation)
                if is_abstract:
                    if name in namespace:
                        if is_class:
                            raise TypeError(
                                f"Abstract class attribute {name} cannot have value"
                            )
                        else:
                            raise TypeError(
                                f"Abstract attribute {name} cannot have value"
                            )
                    if is_class:
                        abstract_class_vars[name] = annotation
                    else:
                        abstract_vars[name] = annotation
        cls.__abstractvars__ = abstract_vars  # pyright: ignore
        cls.__abstractclassvars__ = abstract_class_vars  # pyright: ignore
        return cls

    def __call__(cls, *args, **kwargs):
        if len(cls.__abstractclassvars__) > 0:  # pyright: ignore
            abstract_class_vars = set(cls.__abstractclassvars__)  # pyright: ignore
            raise TypeError(
                f"Can't instantiate abstract class {cls.__name__} with abstract class "
                f"attributes {abstract_class_vars}"
            )
        self = super().__call__(*args, **kwargs)
        abstract_vars = set()
        for name in cls.__abstractvars__:  # pyright: ignore
            # Deliberately not doing `if name in self.__dict__` to allow for use of
            # properties (which are actually class attributes) to override abstract
            # instance variables.
            if getattr(self, name, _sentinel) is _sentinel:
                abstract_vars.add(name)
        if len(abstract_vars) > 0:
            raise TypeError(
                f"Can't instantiate abstract class {cls.__name__} with abstract "
                f"attributes {abstract_vars}"
            )
        return self


@dataclass_transform()
def dataclass(**kwargs):
    def make_dataclass(cls):
        try:
            annotations = cls.__dict__["__annotations__"]
        except KeyError:
            cls = dataclasses.dataclass(**kwargs)(cls)
        else:
            new_annotations = dict(annotations)
            for name, annotation in annotations.items():
                is_abstract, _ = _process_annotation(annotation)
                if is_abstract:
                    new_annotations.pop(name)
            cls.__annotations__ = new_annotations
            cls = dataclasses.dataclass(**kwargs)(cls)
            cls.__annotations__ = annotations
        return cls

    return make_dataclass



import jax.numpy as jnp
from jaxtyping import Array


def left_broadcast_to(arr: Array, shape: tuple[int, ...]) -> Array:
    arr = arr.reshape(arr.shape + (1,) * (len(shape) - arr.ndim))
    return jnp.broadcast_to(arr, shape)



import importlib.metadata

from . import internal as internal, nn as nn
from ._ad import (
    filter_closure_convert as filter_closure_convert,
    filter_custom_jvp as filter_custom_jvp,
    filter_custom_vjp as filter_custom_vjp,
    filter_grad as filter_grad,
    filter_jvp as filter_jvp,
    filter_value_and_grad as filter_value_and_grad,
    filter_vjp as filter_vjp,
)
from ._better_abstract import (
    AbstractClassVar as AbstractClassVar,
    AbstractVar as AbstractVar,
)
from ._caches import clear_caches as clear_caches
from ._callback import filter_pure_callback as filter_pure_callback
from ._enum import Enumeration as Enumeration
from ._eval_shape import filter_eval_shape as filter_eval_shape
from ._filters import (
    combine as combine,
    filter as filter,
    is_array as is_array,
    is_array_like as is_array_like,
    is_inexact_array as is_inexact_array,
    is_inexact_array_like as is_inexact_array_like,
    partition as partition,
)
from ._jit import filter_jit as filter_jit
from ._make_jaxpr import filter_make_jaxpr as filter_make_jaxpr
from ._module import (
    field as field,
    Module as Module,
    module_update_wrapper as module_update_wrapper,
    Partial as Partial,
    static_field as static_field,
)
from ._pretty_print import tree_pformat as tree_pformat, tree_pprint as tree_pprint
from ._serialisation import (
    default_deserialise_filter_spec as default_deserialise_filter_spec,
    default_serialise_filter_spec as default_serialise_filter_spec,
    tree_deserialise_leaves as tree_deserialise_leaves,
    tree_serialise_leaves as tree_serialise_leaves,
)
from ._tree import (
    tree_at as tree_at,
    tree_check as tree_check,
    tree_equal as tree_equal,
    tree_flatten_one_level as tree_flatten_one_level,
    tree_inference as tree_inference,
)
from ._update import apply_updates as apply_updates
from ._vmap_pmap import (
    filter_pmap as filter_pmap,
    filter_vmap as filter_vmap,
    if_array as if_array,
)


__version__ = importlib.metadata.version("equinox")
